# -*- coding: utf-8 -*-
# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the StashAttributes processor
"""
from pickle import FALSE

import pytest
from vermouth.tests.datafiles import (
    FF_UNIVERSAL_TEST,
)
from .helper_functions import create_sys_all_attrs
from vermouth.system import System
from vermouth.molecule import Molecule
from vermouth.forcefield import ForceField
from vermouth.processors.stash_attributes import stash_attributes, StashAttributes


@pytest.fixture
def example_mol():
    mol = Molecule(force_field=ForceField(FF_UNIVERSAL_TEST))
    nodes = [
        {'chain': 'A', 'resname': 'GLY', 'resid': 1},  # 0,
        {'chain': 'A', 'resname': 'GLY', 'resid': 2},  # 1,
        {'chain': 'A', 'resname': 'GLY', 'resid': 3},  # 2,
        {'chain': 'A', 'resname': 'PHE', 'resid': 4},  # 3,
        {'chain': 'B', 'resname': 'GLY', 'resid': 1},  # 4,
        {'chain': 'B', 'resname': 'GLY', 'resid': 2},  # 5,
        {'chain': 'B', 'resname': 'GLY', 'resid': 3},  # 6,
        {'chain': 'B', 'resname': 'PHE', 'resid': 4},  # 7,
        {'chain': 'A', 'resname': 'CYS', 'resid': 5},  # 8,
        {'chain': 'C', 'resname': 'not_protein', 'resid': 1}, # 9
    ]
    mol.add_nodes_from(enumerate(nodes))
    mol.add_edges_from([(0, 1), (1, 2), (2, 3), (4, 5), (5, 6), (6, 7), (3, 8)])
    return mol

@pytest.mark.parametrize('attributes',
                         (('resid',),
                          ('resid', 'chain')))
def test_stash_attributes(example_mol, attributes):
    stash_attributes(example_mol, attributes)

    for node in example_mol.nodes:
        assert len(example_mol.nodes[node]['stash']) == len(attributes)
        for attr in attributes:
            assert example_mol.nodes[node]['stash'].get(attr) == example_mol.nodes[node].get(attr)


@pytest.mark.parametrize('do_twice, expected',
                         ((False, False),
                          (True, True)
                          ))
def test_already_stored(example_mol, caplog, do_twice, expected):
    stash_attributes(example_mol, ('resid',))

    if do_twice:
        stash_attributes(example_mol, ('resid',))

    assert any([rec.levelname == 'WARNING' for rec in caplog.records]) == expected

def test_StashAttributes(example_mol):

    system = System()
    system.molecules.append(example_mol)

    processor = StashAttributes(attributes=('resid',))
    processor.run_system(system)

    for molecule in system.molecules:
        for node in molecule.nodes:
            assert molecule.nodes[node]['stash']['resid'] == molecule.nodes[node]['resid']




# Copyright 2022 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test for the tune idp bonds processor.
"""
import pytest
from vermouth.dssp import dssp
from vermouth.processors.annotate_idrs import AnnotateIDRs, parse_residues
from vermouth.tests.helper_functions import create_sys_all_attrs, test_molecule

@pytest.mark.parametrize('idr_regions, expected', [
    (
    ["1:2"],
    [True, True, True, True, True, False, False, False, False]
    ),
    (
    [],
    [False, False, False, False, False, False, False, False, False]
    )
])
def test_make_disorder_string(test_molecule,
                              idr_regions,
                              expected):
    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1",
              6: "P1", 7: "SN3a", 8: "SP4"}
    # the molecule resnames
    resnames = {0: "A", 1: "A", 2: "A",
                3: "B", 4: "B",
                5: "C",
                6: "D", 7: "D", 8: "D"}
    secstruc = {1: "H", 2: "H", 3: "H", 4: "H"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"resname": resnames,
                                         "atype": atypes})

    AnnotateIDRs(id_regions=idr_regions).run_system(system)
    result = []
    for key, node in system.molecules[0].nodes.items():
        if system.molecules[0].nodes[key].get("cgidr"):
            result.append(True)
        else:
            result.append(False)
    assert result == expected

@pytest.mark.parametrize('idr_regions, secstruc, write_sec, expected',(
        (["1:4"],
         {1: "H", 2: "H", 3: "H", 4: "H"},
         True,
         [{0: "C", 1: "C", 2: "C",
          3: "C", 4: "C",
          5: "C",
          6: "C", 7: "C", 8: "C"}, True]),
        (["1:2"],
         {1: "H", 2: "H", 3: "H", 4: "H"},
         True,
         [{0: "C", 1: "C", 2: "C",
          3: "C", 4: "C",
          5: "H",
          6: "H", 7: "H", 8: "H"}, True]),
        (["1:2"],
         {1: "H", 2: "H", 3: "H", 4: "H"},
         False,
         [{0: None, 1: None, 2: None,
          3: None, 4: None,
          5: None,
          6: None, 7: None, 8: None}, False]),
        (["1:2"],
         {1: "C", 2: "C", 3: "C", 4: "C"},
         True,
         [{0: "C", 1: "C", 2: "C",
          3: "C", 4: "C",
          5: "C",
          6: "C", 7: "C", 8: "C"}, False]),

))
def test_ss_reassign(test_molecule, idr_regions, secstruc, write_sec, expected):
    resnames = {0: "A", 1: "A", 2: "A",
                3: "B", 4: "B",
                5: "C",
                6: "D", 7: "D", 8: "D"}
    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1",
              6: "P1", 7: "SN3a", 8: "SP4"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"resname": resnames,
                                         "atype": atypes},
                                  write_secstruct=write_sec)

    AnnotateIDRs(id_regions=idr_regions).run_system(system)

    for key, node in system.molecules[0].nodes.items():
        assert system.molecules[0].nodes[key].get("cgsecstruct", None) == expected[0][key]

    assert system.molecules[0].meta.get("modified_cgsecstruct", False) == expected[1]

@pytest.mark.parametrize('modify, expected',
                         ((True, True),
                         (False, False)
))
def test_gmx_system_header_supplementary(test_molecule, modify, expected):

    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1",
              6: "P1", 7: "SN3a", 8: "SP4"}
    resnames = {0: "ALA", 1: "ALA", 2: "ALA",
                3: "GLY", 4: "GLY",
                5: "MET",
                6: "ARG", 7: "ARG", 8: "ARG"}
    secstruc ={1: "H", 2: "H", 3: "H", 4: "H"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"resname": resnames,
                                         "atype": atypes})
    if modify:
        AnnotateIDRs(id_regions=["1:2"]).run_system(system)

    dssp.AnnotateResidues(attribute="aasecstruct",
                          sequence="HHHH").run_system(system)
    dssp.AnnotateMartiniSecondaryStructures().run_system(system)

    assert expected == any(["IDR" in i for i in system.meta.get('header', [''])])

@pytest.mark.parametrize('resspec, expected',
                         ((['A-10:20'],
                           [{'chain': 'A', 'resids': [(10, 20)]}]),
                          (['10:20'],
                           [{'chain': None, 'resids': [(10, 20)]}]),
                         (['10:20', 'A-50:65'],
                          [{'chain': None, 'resids': [(10, 20)]}, {'chain': 'A', 'resids': [(50, 65)]}])

                          ))
def test_parse_disorder_resspec(resspec, expected):
    parsed = []
    for spec in resspec:
        parsed.append(parse_residues(spec))
    assert len(parsed) == len(expected)

    for i,j in zip(parsed, expected):
        for key in i.keys():
            assert i[key] == j[key]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides a processor that stores attributes 
"""

from .processor import Processor
from ..log_helpers import StyleAdapter, get_logger
import copy
LOGGER = StyleAdapter(get_logger(__name__))


def stash_attributes(molecule, attributes, stash_name="stash"):
    """
    For each node in molecule, add the attributes to a stash dictionary

    Parameters
    ----------
    molecule: :class:`~vermouth.molecule.Molecule`
        The molecule to transform.
    attributes: tuple[str]
        Attributes to store in the nodes that may otherwise be modified
    stash_name: str
        Name of top level node dictionary to store attributes to
    """
    for node in molecule.nodes:
        for attr in attributes:
            # create the stash if not already there
            stash = molecule.nodes[node].get(stash_name, {})
            molecule.nodes[node][stash_name] = stash
            # stash the attribute if it hasn't already been. Otherwise raise warning.
            if attr not in stash:
                stash[attr] = copy.deepcopy(molecule.nodes[node].get(attr, None))
            else:
                LOGGER.warning("Trying to stash already stashed attribute {} to molecule. Will not stash.", attr)


class StashAttributes(Processor):
    """
    Processor for storing current attributes of a node in a new "stash" attribute

    attributes: tuple[str]
        Attributes to be stashed for later use.
    stash_name: str
        Name of top level node dictionary in which to store existing attributes
    """
    def __init__(self, attributes=(), stash_name='stash'):
        self.attributes = attributes
        self.stash_name = stash_name

    def run_molecule(self, molecule):
        stash_attributes(
            molecule,
            self.attributes,
            self.stash_name
        )
        return molecule

    def run_system(self, system):
        super().run_system(system)



# -*- coding: utf-8 -*-
# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides functionality to read CIF files.
"""

try:
    from CifFile import ReadCif
except ImportError:
    HAVE_READCIF = False
else:
    HAVE_READCIF= True

import numpy as np

from ..molecule import Molecule
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


def casting(value, typeto):
    """
    Cast a value to its correct type with an exception
    value: value to cast
    type: the expected output type of the value
    """
    try:
        return typeto(value)
    # except a ValueError. The two known ones are resid = '.' or charge = '?' which must be a str.
    # may cause an error if more are added
    except ValueError:
        return None

def _cell(cf, entry):
    """
    get the cell dimensions from a cif file as a list

    Parameters
    ----------
    cf: CifFile.CifFile_module.CifFile
        CIF file object parsed by PyCifRW
    entry: str
        Key of the cf dict

    Returns
    -------
    dims: np.array
        array of len(6) containing the dimensions of the cell
    """
    try:
        a = cf[entry]['_cell.length_a']
        b = cf[entry]['_cell.length_b']
        c = cf[entry]['_cell.length_c']
        alpha = cf[entry]['_cell.angle_alpha']
        beta = cf[entry]['_cell.angle_beta']
        gamma = cf[entry]['_cell.angle_gamma']
        dims = np.array([a, b, c, alpha, beta, gamma], dtype=float)
    except KeyError:
        LOGGER.info("_cell information missing from .cif file. Will write default dimensions")
        dims = np.array([1, 1, 1, 90, 90, 90])

    return dims


def cif_entry_reader(cf, entry,
                     cif_categories_all, cif_category_names, cif_category_types, essential_properties,
                     modelidx, exclude, ignh):
    """
    Read a single entry of a parsed cif file

    Parameters
    ----------
    cf: :func: `<CifFile.CifFile_module.CifFile>`
        CIF file object parsed by PyCifRW
    entry: str
        Key of the cf dict
    cif_categories_all: list
        List of valid category keys contained within a cif file entry
    cif_category_names: list
        Common name for category entries
    cif_category_types: list
        Type to cast CIF data to
    essential_properties: list
        List of properties which must be contained
    modelidx: int
        If the cif file contains multiple models, which one to select.
    exclude: collections.abc.Container[str]
        Atoms that have one of these residue names will not be included.
    ignh: bool
        Whether hydrogen atoms should be ignored.

    Returns
    -------
    vermouth.molecule.Molecule
        A parsed molecule. Will not contain edges

    """
    # first filter the data by which categories are present
    # make list of the category names which are present
    names = []
    # make a list of the lists of the data
    all_data = []
    for category, name, cattype in zip(cif_categories_all, cif_category_names, cif_category_types):
        if category in cf[entry]:
            # cast the data to its correct type
            data = [casting(i, cattype) for i in cf[entry][category]]
            all_data.append(data)
            names.append(name)
        # if we're missing one of the essential categories, raise a warning
        elif name in essential_properties:
            LOGGER.warning(f"{name} data is missing from the input file, and is a required field.")

    # if we don't have atomnames but do have element information, copy it.
    if ('atomname' not in names) and ('element' in names):
        LOGGER.warning("atomname data missing from input file. Will attempt to continue using element data.")
        names.append('atomname')
        all_data.append([casting(i, str) for i in cf[entry]['_atom_site.type_symbol']])

    # for each atom, make a dictionary with its associated name
    properties_dict_list = [dict(zip(names, v)) for v in zip(*all_data)]

    molecule = Molecule()

    # find the indices in the data which are actually from the model that we're after.
    model_atom_indices = [index for index, value in
                          enumerate([int(model) == modelidx for
                                     model in cf[entry]['_atom_site.pdbx_PDB_model_num']])
                          if value]

    idx = 0  # add nodes by separate index in case we skip some atoms
    for index in model_atom_indices:
        properties = properties_dict_list[index]

        if properties.get('resname', None) in exclude or (ignh and properties['element'] == 'H'):
            continue

        properties['position'] = np.array([properties['x'],
                                           properties['y'],
                                           properties['z'],
                                           ], dtype=float) / 10

        molecule.add_node(idx, **properties)
        idx += 1

    molecule.box = _cell(cf, entry)

    return molecule

def read_cif_file(file_name, exclude=('SOL', 'HOH'), ignh=False, modelidx=1):
    """
    Parse a CIF file to create a list of molecules using the PyCIFRW library

    Parameters
    ----------
    file_name: str
        The file to read.
    exclude: collections.abc.Container[str]
        Atoms that have one of these residue names will not be included.
    ignh: bool
        Whether hydrogen atoms should be ignored.
    modelidx: int
        If the cif file contains multiple models, which one to select.

    Returns
    -------
    list[vermouth.molecule.Molecule]
        The parsed molecules. Will not contain edges
    """

    # list of categories from the read cif file to read into a molecule
    cif_categories_all = [
        '_atom_site.id',  # atomid
        '_atom_site.label_atom_id',  # atomname
        '_atom_site.label_comp_id',  # resname
        '_atom_site.label_asym_id',  # chain
        '_atom_site.label_seq_id',  # resid
        '_atom_site.pdbx_pdb_ins_code',  # insertion_code
        '_atom_site.cartn_x',  # x
        '_atom_site.cartn_y',  # y
        '_atom_site.cartn_z',  # z
        '_atom_site.occupancy',  # occupancy
        '_atom_site.b_iso_or_equiv',  # temp_factor
        '_atom_site.type_symbol',  # element
        '_atom_site.pdbx_formal_charge',  # charge (nb not added by AF3)
        '_atom_site.pdbx_PDB_model_num'  # model
    ]
    cif_category_names = ['atomid', 'atomname', 'resname', 'chain', 'resid', 'insertion_code',
                          'x', 'y', 'z',
                          'occupancy', 'temp_factor',
                          'element', 'charge', 'model']
    cif_category_types = [int, str, str, str, int, str,
                          float, float, float,
                          float, float,
                          str, float, int]
    essential_properties = ['resname', 'resid']

    cf = ReadCif(str(file_name))
    # PyCifRW seems to store everything from the file it reads in a top level key from _entry.id
    # which ~ corresponds to the pdb code. From a single file we should only have one key.
    if len(cf.keys()) > 1:
        LOGGER.info("This cif file contains multiple entries. Check output carefully.")

    molecules = []
    for entry in cf.keys():
        molecule = cif_entry_reader(cf, entry, cif_categories_all,
                                    cif_category_names, cif_category_types, essential_properties,
                                    modelidx, exclude, ignh)
        molecules.append(molecule)

    return molecules

class CIFReader():
    def __init__(self, file, exclude, ignh, modelidx=1):
        self.input_file = file
        self.exclude = exclude
        self.ignh = ignh
        self.modelidx = modelidx

    def reader(self):
        if HAVE_READCIF:
            molecules = read_cif_file(self.input_file, self.exclude, self.ignh, self.modelidx)
            return molecules
        else:
            raise ImportError("The PyCifRW library must be installed to read .cif files with Vermouth.")




# -*- coding: utf-8 -*-
# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unittests for the CIF reader.
"""

import numpy as np
import pytest
import logging
import sys
from CifFile import ReadCif
import os.path
import subprocess

from vermouth.forcefield import ForceField
from vermouth.gmx import read_itp
import vermouth.pdb.cif as cif
from vermouth.tests.datafiles import (CIF_PROTEIN, CIF_NO_CELL, CIF_PDB_COMPARE, CIF_MULTI, CIF_MISSING_RESNAME,
                                      CIF_MISSING_ATOMNAME)
from ..helper_functions import find_in_path
from ..integration_tests.test_integration import compare_pdb, assert_equal_blocks


@pytest.mark.parametrize('input, expected',
                         (
                                 (CIF_PROTEIN, np.array([50.840, 42.770, 28.950, 90, 90, 90])),
                                 (CIF_NO_CELL, np.array([1, 1, 1, 90, 90, 90])),

                         )
                         )
def test_cell(input, expected):
    input_data = ReadCif(str(input))
    model_name = '1UBQ'

    assert cif._cell(input_data, model_name) == pytest.approx(expected)


def cif_lines():
    """
    lines from CIF_PROTEIN of the first two residues
    """
    cif = '''ATOM   1   N N   . MET A 1 1  ? 27.340 24.430 2.614  1.00 9.67  ? 1   MET A N   1
ATOM   2   C CA  . MET A 1 1  ? 26.266 25.413 2.842  1.00 10.38 ? 1   MET A CA  1
ATOM   3   C C   . MET A 1 1  ? 26.913 26.639 3.531  1.00 9.62  ? 1   MET A C   1
ATOM   4   O O   . MET A 1 1  ? 27.886 26.463 4.263  1.00 9.62  ? 1   MET A O   1
ATOM   5   C CB  . MET A 1 1  ? 25.112 24.880 3.649  1.00 13.77 ? 1   MET A CB  1
ATOM   6   C CG  . MET A 1 1  ? 25.353 24.860 5.134  1.00 16.29 ? 1   MET A CG  1
ATOM   7   S SD  . MET A 1 1  ? 23.930 23.959 5.904  1.00 17.17 ? 1   MET A SD  1
ATOM   8   C CE  . MET A 1 1  ? 24.447 23.984 7.620  1.00 16.11 ? 1   MET A CE  1
ATOM   9   N N   . GLN A 1 2  ? 26.335 27.770 3.258  1.00 9.27  ? 2   GLN A N   1
ATOM   10  C CA  . GLN A 1 2  ? 26.850 29.021 3.898  1.00 9.07  ? 2   GLN A CA  1
ATOM   11  C C   . GLN A 1 2  ? 26.100 29.253 5.202  1.00 8.72  ? 2   GLN A C   1
ATOM   12  O O   . GLN A 1 2  ? 24.865 29.024 5.330  1.00 8.22  ? 2   GLN A O   1
ATOM   13  C CB  . GLN A 1 2  ? 26.733 30.148 2.905  1.00 14.46 ? 2   GLN A CB  1
ATOM   14  C CG  . GLN A 1 2  ? 26.882 31.546 3.409  1.00 17.01 ? 2   GLN A CG  1
ATOM   15  C CD  . GLN A 1 2  ? 26.786 32.562 2.270  1.00 20.10 ? 2   GLN A CD  1
ATOM   16  O OE1 . GLN A 1 2  ? 27.783 33.160 1.870  1.00 21.89 ? 2   GLN A OE1 1
ATOM   17  N NE2 . GLN A 1 2  ? 25.562 32.733 1.806  1.00 19.49 ? 2   GLN A NE2 1
'''

    data = []
    for line in cif.splitlines():
        tokens = line.split()
        anum = tokens[1]
        aname = tokens[-2]
        resname = tokens[-4]
        chain = tokens[-3]
        resid = tokens[-5]
        insert_code = tokens[9]
        x = tokens[10]
        y = tokens[11]
        z = tokens[12]
        position = np.array([tokens[10], tokens[11], tokens[12]], dtype=float) / 10
        element = tokens[2]
        charge = tokens[-6]
        occupancy = tokens[13]
        temp_factor = tokens[14]
        model = tokens[-1]
        data.append({"atomid": int(anum),
                     "atomname": str(aname),
                     "resname": str(resname),
                     "chain": str(chain),
                     "x": float(x),
                     "y": float(y),
                     "z": float(z),
                     "resid": int(resid),
                     "position": position,
                     "element": str(element),
                     "charge": None,
                     "insertion_code": str(insert_code),
                     "occupancy": float(occupancy),
                     "temp_factor": float(temp_factor),
                     "model": int(model)})

    return data


def test_read_cif_file():
    molecule = cif.read_cif_file(CIF_PROTEIN)[0]

    reference_data = cif_lines()

    for i in sorted(molecule.nodes)[:len(reference_data)]:
        assert all(j in molecule.nodes[i].keys() for j in reference_data[i].keys())

        for key, value in molecule.nodes[i].items():
            assert value == pytest.approx(reference_data[i][key])


def test_CIFReader():
    processor = cif.CIFReader(file=CIF_PROTEIN, exclude=('HOH'), ignh=False)

    molecules = processor.reader()

    assert len(molecules) == 1


def test_equal_output(tmp_path):
    martinize2 = find_in_path()

    command0 = [
        sys.executable,
        martinize2,
        '-f', str(CIF_PROTEIN),
        '-o', 'cif.top',
        '-x', 'cif.pdb',
        '-ignore', 'HOH', '-name', 'cif'
    ]
    command1 = [
        sys.executable,
        martinize2,
        '-f', str(CIF_PDB_COMPARE),
        '-o', 'pdb.top',
        '-x', 'pdb.pdb',
        '-ignore', 'HOH', '-name', 'pdb'
    ]
    # run commands
    proc0 = subprocess.run(command0, cwd=tmp_path, timeout=90, check=False)
    assert proc0.returncode == 0
    proc1 = subprocess.run(command1, cwd=tmp_path, timeout=90, check=False)
    assert proc1.returncode == 0

    # find files and assert only 2 files of each
    itp_files = sorted(os.path.abspath(fname) for fname in tmp_path.glob('*.itp'))
    pdb_files = sorted(os.path.abspath(fname) for fname in tmp_path.glob('*.pdb'))
    assert len(itp_files) == 2
    assert len(pdb_files) == 2

    # compare pdb. easily read in
    compare_pdb(pdb_files[0], pdb_files[1])

    # need to take care with itps because we have to use different names to make sure we don't overwrite in tmp_path.
    cif_itp = [file for file in itp_files if 'cif' in file][0]
    pdb_itp = [file for file in itp_files if 'pdb' in file][0]

    dummy_ff = ForceField(name='dummy')
    with open(cif_itp) as fn1:
        read_itp(fn1, dummy_ff)
    dummy_ff2 = ForceField(name='dummy')
    with open(pdb_itp) as fn2:
        read_itp(fn2, dummy_ff2)

    assert_equal_blocks(dummy_ff.blocks['cif_0'],
                        dummy_ff2.blocks['pdb_0'],
                        blocknames_equal=False)

def test_missing_resnames(caplog):
    cif.read_cif_file(CIF_MISSING_RESNAME)
    assert any([rec.levelname == 'WARNING' for rec in caplog.records])


def test_missing_atomnames(caplog):
    cif.read_cif_file(CIF_MISSING_ATOMNAME)
    assert any([rec.levelname == 'WARNING' for rec in caplog.records])


def test_multiple_entries(caplog):
    caplog.set_level(logging.INFO)
    cif.read_cif_file(CIF_MULTI)
    assert any([rec.levelname == 'INFO' for rec in caplog.records])



# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test some of the behaviour of :class:`vermouth.system.System`
"""

import pytest

from vermouth import System, Molecule
from vermouth.forcefield import ForceField


def test_add_molecule_mismatch_ff():
    """
    Assert that adding a molecule to a system with a mismatching force field
    errors.
    """
    ff_a = ForceField(name='a')
    ff_b = ForceField(name='b')
    assert ff_a != ff_b

    system = System(force_field=ff_a)
    mol_a = Molecule(force_field=ff_a)
    system.add_molecule(mol_a)
    assert len(system.molecules) == 1
    mol_b = Molecule(force_field=ff_b)
    with pytest.raises(KeyError):
        system.add_molecule(mol_b)


def test_system_automatic_ff():
    """
    Assert that a system automatically gets a force field when given a molecule.
    """
    ff_a = ForceField(name='a')
    ff_b = ForceField(name='b')
    assert ff_a != ff_b

    system = System(force_field=None)
    mol_a = Molecule(force_field=ff_a)
    system.add_molecule(mol_a)
    assert len(system.molecules) == 1
    assert system.force_field == ff_a
    mol_b = Molecule(force_field=ff_b)
    with pytest.raises(KeyError):
        system.add_molecule(mol_b)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides processors that can add and remove IDR specific bonds
"""

from itertools import chain
from ..dssp.dssp import SS_CG, sequence_from_residues
from ..selectors import is_protein
from .processor import Processor
from ..rcsu.go_utils import _in_chain_and_resid_region
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


def parse_residues(resspec):
    """
    Parse a residue specification: [<chain>-][<resid_start>]:[<resid_end>] where
    resid is /[0-9]+/.
    Returns a dictionary with keys 'chain', 'resid_start', and 'resid_end' for the
    fields that are specified. Resids will be ints.

    Parameters
    ----------
    resspec: str

    Returns
    -------
    dict

    """
    # <chain>-<resid>
    *chain, resids = resspec.split('-', 1)
    res_start, res_end = resids.split(':', 1)

    out = {}
    if resids:
        out['resids'] = [(int(res_start), int(res_end))]
    if chain:
        out['chain'] = chain[0]
    else:
        out['chain'] = None
    return out

def annotate_disorder(molecule, id_regions, annotation="cgidr"):
    """
    Annotate the disordered regions of the molecule

    molecule: :class:`vermouth.molecule.Molecule`
            the molecule
    idr_regions: list
        dictionaries defining the disordered regions to annotate
    annotation: str
        name of the annotation in the node
    """

    for region in id_regions:
        for key, node in molecule.nodes.items():
            _old_resid = node['stash']['resid']
            chain = node['chain']
            # make sure we have the correct chain and are in the right region. If no chain in region assume single chain system.
            if _in_chain_and_resid_region(region, _old_resid, chain):
                molecule.nodes[key][annotation] = True
                if "cgsecstruct" in molecule.nodes[key] and molecule.nodes[key]["cgsecstruct"] != "C":
                        molecule.nodes[key]["cgsecstruct"] = "C"
                        molecule.meta['modified_cgsecstruct'] = True
            else:
                molecule.nodes[key][annotation] = False




class AnnotateIDRs(Processor):
    """
    Processor to annotate intrinsically disordered regions of a molecule.

    This processor is designed primarily for the work described in the reference
    M3_GO, but is generally applicable for such circumstances where extra
    addition/removals are necessary.

    """

    def __init__(self, id_regions=None):
        """
        Parameters
        ----------
        id_regions:
            regions defining the IDRs
        """
        self.id_regions = []
        for region in id_regions:
            self.id_regions.append(parse_residues(region))

    def run_molecule(self, molecule):
        """
        Assign disordered regions for a single molecule
        """
        annotate_disorder(molecule, self.id_regions)
        return molecule

    def run_system(self, system):
        """
        Assign the water bias of the Go model to file. Biasing
        is always molecule specific i.e. no two different
        vermouth molecules can have the same bias.

        Parameters
        ----------
        system: :class:`vermouth.system.System`
        """
        if not self.id_regions:
            return system
        LOGGER.info("Annotating disordered regions.", type="step")
        super().run_system(system)

        if any([molecule.meta.get('modified_cgsecstruct', False) for molecule in system.molecules]):
            supplementary_ss_seq = list(
                chain(
                    *(
                        sequence_from_residues(molecule, "cgsecstruct")
                        for molecule in system.molecules
                        if is_protein(molecule)
                    )
                )
            )

            LOGGER.info(("Secondary structure assignment changed between dssp and martinize. "
                         "Check files for details."), type="general")
            system.meta["header"].extend((
                "The assigned secondary structure conflicted with ",
                "annotated IDRs. The following sequence of Martini secondary ",
                "structure was actually applied to the system:",
                "".join([SS_CG[i] for i in supplementary_ss_seq])
            ))



# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test the ApplyPosres processor and the related functions.
"""
from vermouth import selectors
from vermouth.processors import apply_posres
from vermouth.tests.helper_functions import test_molecule, create_sys_all_attrs

def test_apply_posres(test_molecule):

    selector =  selectors.select_backbone

    mol = apply_posres.apply_posres(molecule=test_molecule,
                                    selector=selector,
                                    atomnames='BB',
                                    force_constant=1000)

    expected_interaction_sites = [i for i in mol.nodes if selector(mol.nodes[i])]
    posres_atoms = [i.atoms[0] for i in mol.interactions['position_restraints']]

    assert set(expected_interaction_sites) == set(posres_atoms)

def test_ApplyPosres(test_molecule):
    resnames = {0: "ALA", 1: "ALA", 2: "ALA",
                3: "GLY", 4: "GLY",
                5: "MET",
                6: "ARG", 7: "ARG", 8: "ARG"}
    secstruc =  {1: "H", 2: "H", 3: "H", 4: "H"}

    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1",
              6: "P1", 7: "SN3a", 8: "SP4"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"resname": resnames,
                                         "atype": atypes})

    apply_posres.ApplyPosres((selectors.select_backbone, "BB"), 1000).run_system(system)

    assert all(mol.interactions.get('position_restraints') for mol in system.molecules)
    assert all(mol.meta.get('define').get('POSRES_FC') == 1000 for mol in system.molecules)




# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..processors.processor import Processor
import numpy as np
from scipy.spatial.distance import euclidean
from scipy.spatial import cKDTree as KDTree
from ..graph_utils import make_residue_graph
from itertools import product
from vermouth.file_writer import deferred_open
from collections import defaultdict
from vermouth import __version__ as VERSION
from pathlib import Path

# BOND TYPE
# Types of contacts:
# HB -- 1 -- hydrogen-bond
# PH -- 2 -- hydrophobic
# AR -- 3 -- aromatic - contacts between aromatic rings
# IB -- 4 -- ionic bridge - contacts created by two atoms with different charges
# DC -- 5 -- destabilizing contact - contacts which are in general repulsive
# OT -- 6 -- denotes negligible other contacts.
# 1-HB,2-PH,3-AR,4-IP,5-DC,6-OT
BOND_TYPE = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 1, 1, 1, 5, 5, 6, 6, 6, 1, 1],
                      [0, 1, 5, 1, 5, 5, 6, 6, 6, 1, 5],
                      [0, 1, 1, 5, 5, 5, 6, 6, 6, 5, 1],
                      [0, 5, 5, 5, 2, 2, 6, 6, 6, 5, 5],
                      [0, 5, 5, 5, 2, 3, 6, 6, 6, 5, 5],
                      [0, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
                      [0, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
                      [0, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6],
                      [0, 1, 1, 5, 5, 5, 6, 6, 6, 5, 4],
                      [0, 1, 5, 1, 5, 5, 6, 6, 6, 4, 5]])

PROTEIN_MAP = {
    "ALA": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0}
    },
    "ARG": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 4},
        'CD':       {'vrad': 1.88, 'atype': 7},
        'NE':       {'vrad': 1.64, 'atype': 3},
        'CZ':       {'vrad': 1.61, 'atype': 6},
        'NH1':      {'vrad': 1.64, 'atype': 3},
        'NH2':      {'vrad': 1.64, 'atype': 3},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0}
    },
    "ASN": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.61, 'atype': 6},
        'OD1':      {'vrad': 1.42, 'atype': 2},
        'ND2':      {'vrad': 1.64, 'atype': 3},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "ASP": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.61, 'atype': 6},
        'OD1':      {'vrad': 1.46, 'atype': 2},
        'OD2':      {'vrad': 1.42, 'atype': 2},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "CYM": {},
    "CYX": {},
    "CYS": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'SG':       {'vrad': 1.77, 'atype': 6},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "GLN": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 4},
        'CD':       {'vrad': 1.61, 'atype': 6},
        'OE1':      {'vrad': 1.42, 'atype': 2},
        'NE2':      {'vrad': 1.64, 'atype': 3},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "GLU": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 4},
        'CD':       {'vrad': 1.61, 'atype': 6},
        'OE1':      {'vrad': 1.46, 'atype': 2},
        'OE2':      {'vrad': 1.42, 'atype': 2},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "GLY": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':        {'vrad': 1.88, 'atype': 6},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "HIE": {},
    "HIP": {},
    "HIS": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.61, 'atype': 5},
        'ND1':      {'vrad': 1.64, 'atype': 1},
        'CD2':      {'vrad': 1.76, 'atype': 5},
        'CE1':      {'vrad': 1.76, 'atype': 5},
        'NE2':      {'vrad': 1.64, 'atype': 1},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "ILE": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG1':      {'vrad': 1.88, 'atype': 4},
        'CG2':      {'vrad': 1.88, 'atype': 4},
        'CD':       {'vrad': 1.88, 'atype': 4},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "LEU": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 4},
        'CD1':      {'vrad': 1.88, 'atype': 4},
        'CD2':      {'vrad': 1.88, 'atype': 4},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "LYS": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 4},
        'CD':       {'vrad': 1.88, 'atype': 4},
        'CE':       {'vrad': 1.88, 'atype': 7},
        'NZ':       {'vrad': 1.64, 'atype': 3},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "MET": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 4},
        'SD':       {'vrad': 1.77, 'atype': 8},
        'CE':       {'vrad': 1.88, 'atype': 4},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "PHE": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 5},
        'CD1':      {'vrad': 1.61, 'atype': 5},
        'CD2':      {'vrad': 1.76, 'atype': 5},
        'CE1':      {'vrad': 1.76, 'atype': 5},
        'CE2':      {'vrad': 1.76, 'atype': 5},
        'CZ':       {'vrad': 1.76, 'atype': 5},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "PRO": {
        'N':        {'vrad': 1.64, 'atype': 6},
        'CA':       {'vrad': 1.88, 'atype': 4},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.88, 'atype': 4},
        'CD':       {'vrad': 1.88, 'atype': 4},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "SER": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 6},
        'OG':       {'vrad': 1.46, 'atype': 1},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "THR": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 6},
        'OG1':      {'vrad': 1.46, 'atype': 1},
        'CG2':      {'vrad': 1.88, 'atype': 4},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "TRP": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.61, 'atype': 5},
        'CD1':      {'vrad': 1.76, 'atype': 5},
        'CD2':      {'vrad': 1.61, 'atype': 5},
        'NE1':      {'vrad': 1.64, 'atype': 3},
        'CE2':      {'vrad': 1.61, 'atype': 5},
        'CE3':      {'vrad': 1.76, 'atype': 5},
        'CZ2':      {'vrad': 1.76, 'atype': 5},
        'CZ3':      {'vrad': 1.76, 'atype': 5},
        'CH2':      {'vrad': 1.76, 'atype': 5},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "TYR": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG':       {'vrad': 1.61, 'atype': 5},
        'CD1':      {'vrad': 1.76, 'atype': 5},
        'CD2':      {'vrad': 1.76, 'atype': 5},
        'CE1':      {'vrad': 1.76, 'atype': 5},
        'CE2':      {'vrad': 1.76, 'atype': 5},
        'CZ':       {'vrad': 1.61, 'atype': 5},
        'OH':       {'vrad': 1.46, 'atype': 1},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    },
    "VAL": {
        'N':        {'vrad': 1.64, 'atype': 3},
        'CA':       {'vrad': 1.88, 'atype': 7},
        'C':        {'vrad': 1.61, 'atype': 6},
        'O':        {'vrad': 1.42, 'atype': 2},
        'CB':       {'vrad': 1.88, 'atype': 4},
        'CG1':      {'vrad': 1.88, 'atype': 4},
        'CG2':      {'vrad': 1.88, 'atype': 4},
        'OXT':      {'vrad': 1.42, 'atype': 2},
        'default':  {'vrad': 0.00, 'atype': 0},
    }
}


def _get_vdw_radius(resname, atomname):
    """
    get the vdw radius of an atom indexed internally within a serially numbered residue
    """
    try:
        res_vdw = PROTEIN_MAP[resname]
    except KeyError:
        return 0.00

    try:
        atom_vdw = res_vdw[atomname]
    except KeyError:
        atom_vdw = res_vdw['default']
    return atom_vdw['vrad']


def _get_atype(resname, atomname):
    """
    get the vdw radius of an atom indexed internally within a serially numbered residue
    """
    try:
        res_vdw = PROTEIN_MAP[resname]
    except KeyError:
        return 0

    try:
        atom_vdw = res_vdw[atomname]
    except KeyError:
        atom_vdw = res_vdw['default']

    return atom_vdw['atype']


def _make_surface(position, fiba, fibb, vrad):
    """
    Generate points on a sphere using Fibonacci points

    position: np.array
        shape (3,) array of an atomic position to build a sphere around
    fiba: int. n-1 fibonacci number to build number of points on sphere
    fibb: int. n fibonacci number to build number of points on sphere
    vrad: float. VdW radius of the input atom to build a sphere around.

    position: centre of sphere
    """

    x, y, z = position

    k = np.arange(fibb)
    phi_aux = (np.arange(1, fibb+1) * fiba) % fibb
    phi_aux[phi_aux == 0] = fibb
    theta = np.arccos(1.0 - 2.0 * k / fibb)
    phi = 2.0 * np.pi * phi_aux / fibb
    surface_x = x + vrad * np.sin(theta) * np.cos(phi)
    surface_y = y + vrad * np.sin(theta) * np.sin(phi)
    surface_z = z + vrad * np.cos(theta)
    surface = np.stack((surface_x, surface_y, surface_z), axis=-1)

    return surface


def atom2res(arrin, nresidues, atom_map, norm=False):
    '''
    take an array with atom level data and sum the entries over within the residue
    '''

    out = np.zeros((nresidues, nresidues))
    for res_idx, res_jdx in product(atom_map.keys(), atom_map.keys()):
        atom_idxs = atom_map[res_idx]
        atom_jdxs = atom_map[res_jdx][:, np.newaxis]
        value = arrin[atom_idxs, atom_jdxs].sum()
        out[res_idx, res_jdx] = value

    if norm:
        out[out > 0] = 1

    return out


def _contact_info(molecule):
    """
    get the atom attributes that we need to calculate the contacts
    """

    G = make_residue_graph(molecule)

    resids = []
    chains = []
    resnames = []
    positions_all = []
    ca_pos = []
    vdw_list = []
    atypes = []
    res_serial = []
    res_idx = []
    for residue in G.nodes:
        # we only need these for writing at the end
        resnames.append(G.nodes[residue]['resname'])
        resids.append(G.nodes[residue]['resid'])
        chains.append(G.nodes[residue]['chain'])
        res_idx.append(G.nodes[residue]['_res_serial'])
        subgraph = G.nodes[residue]['graph']

        for atom in sorted(subgraph.nodes):
            position = subgraph.nodes[atom].get('position', [np.nan]*3)
            if np.isfinite(position).all():
                res_serial.append(subgraph.nodes[atom]['_res_serial'])

                positions_all.append(subgraph.nodes[atom]['position'] * 10)

                vdw_list.append(_get_vdw_radius(subgraph.nodes[atom]['resname'],
                                                subgraph.nodes[atom]['atomname']))
                atypes.append(_get_atype(subgraph.nodes[atom]['resname'],
                                         subgraph.nodes[atom]['atomname']))

                if subgraph.nodes[atom]['atomname'] == 'CA':
                    ca_pos.append(subgraph.nodes[atom]['position'])


    vdw_list = np.array(vdw_list)
    atypes = np.array(atypes)
    coords = np.stack(positions_all)
    res_serial = np.array(res_serial)

    resids = np.array(resids)
    chains = np.array(chains)
    resnames = np.array(resnames)
    res_idx = np.array(res_idx)

    # 2) find the number of residues that we have
    nresidues = len(G)

    return vdw_list, atypes, coords, res_serial, resids, chains, resnames, res_idx, ca_pos, nresidues, G

def _calculate_overlap(coords_tree, vdw_list, natoms, vdw_max, alpha=1.24):
    """
    Find enlarged (OV) overlap contacts

    coords_tree: KDTree
        KDTree of the input coordinates
    vdw_list: list
        list of vdw radii of the input coordinates
    natoms: int
        number of atoms in the molecule
    vdw_max: float
        maximum possible vdw radius of atoms
    alpha: float
        Enlargement factor for attraction effects
    """
    over = np.zeros((natoms, natoms))
    over_sdm = coords_tree.sparse_distance_matrix(coords_tree, 2 * vdw_max * alpha)
    for (idx, jdx), distance_between in over_sdm.items():
        if idx != jdx:
            if distance_between < (vdw_list[idx] + vdw_list[jdx]) * alpha:
                over[idx, jdx] = 1
    return over

def _calculate_csu(coords, vdw_list, fiba, fibb, natoms, coords_tree, vdw_max, water_radius=2.80):
    """
    Calculate contacts of structural units (CSU)

    coords: Nx3 numpy array
        coordinates of atoms in the molecule
    vdw_list: list
        vdw radii of the atoms in the molecule
    fiba, fibb: int
        n-1th and nth fibonacci numbers from which to generate points on a sphere around the input coordinate
    natoms: int
        number of atoms in the molecule
    coords_tree: KDTree
        KDTree of the input coordinates
    vdw_max: float
        maximum possible vdw radius of atoms
    water_radius: float
        radius of water molecule in A

    Returns:
    hit_results: natoms x fibb np.array
        each i,j entry is the index of the atom in coords which is the closest atom to atom i at index j of the
        fibonacci sphere

    """

    #setup arrays to keep track
    hit_results = np.full((natoms, fibb), -1)
    dists_counter = np.full((natoms, fibb), np.inf)

    # sparse matrix with a cutoff at the maximum possible distance for a contact
    surface_sdm = coords_tree.sparse_distance_matrix(coords_tree, (2 * vdw_max) + water_radius)
    # n.b. this loop works because sparse_distance_matrix is sorted by (idx, jdx) pairs
    for (idx, jdx), distance_between in surface_sdm.items():
        # don't take atoms which are identical
        if idx == jdx:
            continue

        # check that the distance between them is shorter than the vdw sum and the water radius
        if distance_between >= (vdw_list[idx] + vdw_list[jdx] + water_radius):
            continue

        # Generate the fibonacci sphere for this point and make a KDTree from it
        base_tree = KDTree(_make_surface(coords[idx], fiba, fibb, vdw_list[idx]+water_radius))

        # find points on the base point sphere which are within the vdw cutoff of the target point's coordinate
        res = np.array(base_tree.query_ball_point(coords[jdx], vdw_list[jdx] + water_radius))

        # if we have any results
        if len(res) > 0:
            # find where the distance between the two points is smaller than the current recorded distance
            # at the points which are within the cutoff
            to_fill = np.where(distance_between < dists_counter[idx][res])[0]

            # record the new distances and indices of the points
            dists_counter[idx][res[to_fill]] = distance_between
            hit_results[idx][res[to_fill]] = jdx

    return hit_results


def _contact_types(hit_results, natoms, atypes):
    """
    From CSU contacts, establish contact types from atomtypes

    hit_results: NxM ndarray
        array for N atoms in molecule for M fibonnaci points on each atom.
        Each i,j entry is the index of the atom which is the closest contact to i
    natoms: int
        number of atoms in the molecule
    atypes: array
        list of the atomtypes of each atom in the molecule
    """

    contactcounter_1 = np.zeros((natoms, natoms))
    stabilisercounter_1 = np.zeros((natoms, natoms))
    destabilisercounter_1 = np.zeros((natoms, natoms))

    for i, j in enumerate(hit_results):
        for k in j:
            if k >= 0:
                at1 = atypes[i]
                at2 = atypes[k]
                if (at1 > 0) and (at2 > 0):
                    contactcounter_1[i, k] += 1
                    btype = BOND_TYPE[at1, at2]
                    if btype <= 4:
                        stabilisercounter_1[i, k] += 1
                    if btype == 5:
                        destabilisercounter_1[i, k] += 1

    return contactcounter_1, stabilisercounter_1, destabilisercounter_1

def make_atom_map(res_serial):

    atom_map = defaultdict(list)
    for atom_idx, res_idx in enumerate(res_serial):
        atom_map[res_idx].append(atom_idx)
    for key, value in atom_map.items():
        atom_map[key] = np.array(value)

    return atom_map

def _calculate_contacts(vdw_list, atypes, coords, res_serial, nresidues):
    """
    run the contact calculation functions

    vdw_list: np.array
        list of the vdw radii of the atoms in the system
    atypes: np.array
        list of the atom types in the system to determine the nature of contacts
    coords: nx3 array
        coordinates of all the atoms in the system
    res_serial: np.array
        list of the serial residue number of each atom in the system
    nresidues: int
        number of residues in the system
    """

    # some initial definitions of variables that we need
    fib = 14
    fiba, fibb = 0, 1
    for _ in range(fib):
        fiba, fibb = fibb, fiba + fibb

    natoms = len(coords)

    vdw_max = max(item['vrad'] for atoms in PROTEIN_MAP.values() for item in atoms.values())

    # make the KDTree of the input coordinates
    coords_tree = KDTree(coords)

    # calculate the OV contacts of the molecule
    over = _calculate_overlap(coords_tree, vdw_list, natoms, vdw_max, alpha=1.24)

    # Calculate the CSU contacts of the molecule
    hit_results = _calculate_csu(coords, vdw_list, fiba, fibb, natoms, coords_tree, vdw_max, water_radius=2.80)

    # find the types of contacts we have
    contactcounter_1, stabilisercounter_1, destabilisercounter_1 = _contact_types(hit_results, natoms, atypes)

    atom_map = make_atom_map(res_serial)

    # transform the resolution between atoms and residues
    overlapcounter_2 = atom2res(over, nresidues, atom_map, norm=True)
    contactcounter_2 = atom2res(contactcounter_1, nresidues, atom_map)
    stabilisercounter_2 = atom2res(stabilisercounter_1, nresidues, atom_map)
    destabilisercounter_2 = atom2res(destabilisercounter_1, nresidues, atom_map)

    return overlapcounter_2, contactcounter_2, stabilisercounter_2, destabilisercounter_2


def _get_contacts(nresidues, overlaps, contacts, stabilisers, destabilisers, res_idx, G):
    '''
    Generate contacts list from the contact arrays calculated

    nresidues: int
        number of residues in the molecule
    overlaps: ndarray
        nresidues x nresidues array of OV contacts in the molecule
    contacts: ndarray
        nresidues x nresidues array of CSU contacts in the molecule
    stabilisers: ndarray
        nresidues x nresidues array of CSU stabilising contacts in the molecule
    destabilisers: ndarray
        nresidues x nresidues array of CSU destabilising contacts in the molecule
    res_idx: list
        list of serial residue ids for each of the residues
    G: nx.Graph
        residue based graph of the molecule
    '''
    contacts_list = []
    all_contacts = []
    for i1, i2 in product(np.arange(nresidues), np.arange(nresidues)):
        if i1 == i2:
            continue
        over = overlaps[i1, i2]
        cont = contacts[i1, i2]
        stab = stabilisers[i1, i2]
        dest = destabilisers[i1, i2]
        rcsu = (stab - dest) > 0

        if (over > 0 or cont > 0):
            a = np.where(res_idx == i1)[0][0]
            b = np.where(res_idx == i2)[0][0]
            all_contacts.append([i1+1, i2+1, a, b, over, cont, stab, rcsu])
            if over == 1 or (over == 0 and rcsu):
                # this is a OV or rCSU contact we take it
                contacts_list.append((int(G.nodes[a]['stash']['resid']), G.nodes[a]['chain'],
                                      int(G.nodes[b]['stash']['resid']), G.nodes[b]['chain']))

    return contacts_list, all_contacts


def _write_contacts(fout, all_contacts, ca_pos, G):
    '''
    write the contacts calculated to file
    fout: str
        path to write file to
    all_contacts: list
        list of lists of every contact found
    ca_pos: list
        list of (3,) arrays with the position of the CA atom of each residue
    G: nx.Graph
        residue graph of the input molecule
    '''

    header = [f"Go contact map calculated with vermouth {VERSION}\n\n"]

    header.append("Residue-Residue Contacts\n"
                  "\n"
                  "ID       - atom identification\n"
                  "I1,I2    - serial residue id\n"
                  "AA       - 3-letter code of aminoacid\n"
                  "C        - chain\n"
                  "I(PDB)   - residue number in PDB file\n"
                  "DCA      - distance between CA\n"
                  "CMs      - OV , CSU , oCSU , rCSU\n"
                  "           (CSU does not take into account chemical properties of atoms)\n"
                  "rCSU     - net contact from rCSU\n"
                  "Count    - number of contacts between residues\n"
                  "\n"
                  "      ID    I1  AA  C I(PDB)     I2  AA  C I(PDB)        DCA       CMs    rCSU   Count \n"
                  "=======================================================================================\n")

    msgs = []
    count = 0
    for contact in all_contacts:
        count += 1
        msg = (f"R {int(count):6d} "
               f"{int(contact[0]):5d}  {G.nodes[contact[2]]['resname']:3s} "
               f"{G.nodes[contact[2]]['chain']:1s} {int(G.nodes[contact[2]]['stash']['resid']):4d}    "
               f"{int(contact[1]):5d}  {G.nodes[contact[3]]['resname']:3s} "
               f"{G.nodes[contact[3]]['chain']:1s} {int(G.nodes[contact[3]]['stash']['resid']):4d}    "
               f"{euclidean(ca_pos[contact[2]], ca_pos[contact[3]])*10:9.4f}     "
               f"{int(contact[4]):1d} {1 if contact[5] != 0 else 0} "
               f"{1 if contact[6] != 0 else 0} {1 if contact[7] else 0}"
               f"{int(contact[7]): 6d}  {int(contact[5]): 6d}\n")
        msgs.append(msg)
    message_out = ''.join(msgs)
    with deferred_open(fout, "w") as f:
        f.write(''.join(header))
        f.write(message_out)


"""
Read RCSU Go model contact maps.
"""


def read_go_map(system, file_path):
    """
    Read a RCSU contact map from the c code as published in
    doi:10.5281/zenodo.3817447. The format requires all
    contacts to have 18 columns and the first column to be
    a capital R.

    Parameters
    ----------
    system: vermouth.system.System
        The system to process. Is modified in-place.
    file_path: :class:`pathlib.Path`
        path to the contact map file

    Returns
    -------
    list(tuple)
        contact as chain id, res id, chain id, res id
    """
    with open(file_path, "r", encoding='UTF-8') as _file:
        contacts = []
        for line in _file:
            tokens = line.strip().split()
            if len(tokens) == 0:
                continue

            if tokens[0] == "R" and len(tokens) == 18:
                # this is a bad place to filter but follows
                # the old script
                if tokens[11] == "1" or (tokens[11] == "0" and tokens[14] == "1"):
                    # this is a OV or rCSU contact we take it
                    contacts.append((int(tokens[5]), tokens[4], int(tokens[9]), tokens[8]))

        if len(contacts) == 0:
            raise IOError("You contact map is empty. Are you sure it has the right formatting?")

    system.go_params["go_map"].append(contacts)


def do_contacts(molecule, write_file):
    '''
    master function to calculate Go contacts

    molecule: vermouth.Molecule
        molecule to calculate contacts for
    write_file: bool
        write the file of the contacts out
    '''
    vdw_list, atypes, coords, res_serial, resids, chains, resnames, res_idx, ca_pos, nresidues, mol_graph = _contact_info(
        molecule)

    overlaps, contacts, stabilisers, destabilisers = _calculate_contacts(vdw_list,
                                                                        atypes,
                                                                        coords,
                                                                        res_serial,
                                                                        nresidues)

    contacts, all_contacts = _get_contacts(nresidues,
                            overlaps, contacts,
                            stabilisers,
                            destabilisers,
                            res_idx,
                            mol_graph)

    if isinstance(write_file, (str, Path)):
        _write_contacts(write_file, all_contacts, ca_pos, mol_graph)

    return contacts


class GenerateContactMap(Processor):
    """
    Processor to generate the contact rCSU contact map for a protein from an atomistic structure
    """
    def __init__(self, write_file):
        self.write_file = write_file

    def run_molecule(self, molecule):
        """
        Process `system`.

        Parameters
        ----------
        system: vermouth.system.System
            The system to process. Is modified in-place.
        """
        return do_contacts(molecule, self.write_file)

    def run_system(self, system):
        for molecule in system.molecules:
            contacts = self.run_molecule(molecule)
            system.go_params["go_map"].append(contacts)



# -*- coding: utf-8 -*-
# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unittests for the CIF reader.
"""

import vermouth
from vermouth import CIFInput
from vermouth.tests.datafiles import CIF_PROTEIN

def test_CIFInput():

    system = vermouth.System()
    CIFInput(filename=CIF_PROTEIN, exclude='HOH', ignh=False, modelidx=1).run_system(system)

    assert len(system.molecules) == 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads quotes, and produces a random one.
"""

import random

from .processor import Processor
from ..data import QUOTES
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


class Quoter(Processor):
    """
    Processor that can produce random string taken from a list. Useful for e.g.
    quotes.

    Parameters
    ----------
    quotes: list[str]
        List of strings describing the quotes.
    """
    def __init__(self, quotes=None):
        if not quotes:
            quotes = QUOTES
        self.quotes = quotes

    def run_system(self, system):
        """
        Logs a random line from the list passed at initialization.

        Parameters
        ----------
        system
            Not used

        Returns
        -------
        None
        """
        LOGGER.info(random.choice(self.quotes))



# -*- coding: utf-8 -*-
# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the Go contact map generator.
"""

import pytest
import numpy as np
from vermouth.rcsu import contact_map
from collections import defaultdict
from vermouth.graph_utils import make_residue_graph
from scipy.spatial import cKDTree as KDTree
from vermouth.rcsu.contact_map import GenerateContactMap
from vermouth.tests.helper_functions import test_molecule, equal_graphs, create_sys_all_attrs
from vermouth.tests.datafiles import TEST_MOLECULE_CONTACT_MAP
from vermouth.file_writer import DeferredFileWriter
from pathlib import Path

@pytest.mark.parametrize('resname, atomname, expected',
                         (
                                 ('TRP', 'N', 1.64),
                                 ('TRP', 'NON', 0.00),
                                 ('NON', 'N', 0.00),
                                 ('NON', 'NON', 0.00)
                         ))
def test_get_vdw_radius(resname, atomname, expected):
    # test that we get the correct vdw radii and errors in resname/atomname are handled correctly
    result = contact_map._get_vdw_radius(resname, atomname)
    assert result == expected


@pytest.mark.parametrize('resname, atomname, expected',
                         (
                                 ('TRP', 'N', 3),
                                 ('TRP', 'NON', 0),
                                 ('NON', 'N', 0),
                                 ('NON', 'NON', 0)
                         ))
def test_get_atype(resname, atomname, expected):
    result = contact_map._get_atype(resname, atomname)
    assert result == expected

def test_surface_generation():
    # test that a surface is generated with the correct points

    position = np.array([1, 1, 1])

    surface = contact_map._make_surface(position, 13, 21, 1)

    assert len(surface) == 21

    first_point = [1.        , 1.        , 2.        ]
    last_point = [1.42591771, 1.        , 0.0952381 ]

    test_points = [first_point, last_point]

    for i, j in enumerate([surface[0], surface[-1]]):
        assert j == pytest.approx(test_points[i])


@pytest.mark.parametrize('norm, expected', (
     (False,
       np.array([[1, 0, 0, 0, 0],
                 [0, 8, 0, 0, 0],
                 [0, 0, 27, 0, 0],
                 [0, 0, 0, 36, 0],
                 [0, 0, 0, 0, 80]]
                )
      ),
     (True,
      np.array([[1, 0, 0, 0, 0],
                [0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 0, 1]]
               )
      )
    ))
def test_atom2res(norm, expected):
    # test that atomic resolution arrays get mapped correctly to their residues and are normalised if required

    arrin = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 3, 3, 3, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 3, 3, 3, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 3, 3, 3, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 4, 4, 4, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 4, 4, 4, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 4, 4, 4, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5],
                      [0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5],
                      [0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5],
                      [0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5],
                      ])

    nres = 5
    res_map = contact_map.make_atom_map([0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4])
    res_array = contact_map.atom2res(arrin, nres, res_map, norm)

    assert np.allclose(res_array, expected)

def test_make_atom_map():
    # test that the atom_map defaultdict is generated correctly

    input = [0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4]
    result = contact_map.make_atom_map(input)

    expected = defaultdict(list, {0: np.array([0]),
                                  1: np.array([1, 2]),
                                  2: np.array([3, 4, 5]),
                                  3: np.array([6, 7, 8]),
                                  4: np.array([9, 10, 11, 12])})

    for key, value in result.items():
        assert key in expected.keys()


def test_contact_info(test_molecule):
    # test we get the expected input data from a molecule

    result = contact_map._contact_info(test_molecule)

    vdw_list, atypes, coords, res_serial, resids, chains, resnames, res_idx, ca_pos, nresidues, G = result

    assert nresidues == 4

    assert np.allclose(vdw_list, [0]*len(test_molecule))
    assert np.allclose(atypes, [0]*len(test_molecule))

    assert np.allclose(coords, np.stack([test_molecule.nodes[node]['position']*10 for node in sorted(test_molecule.nodes)]))

    assert np.allclose(res_serial, np.array([0, 0, 0, 1, 1, 2, 3, 3, 3]))
    assert np.allclose(resids, np.array([1, 2, 3, 4]))
    assert np.allclose(res_idx, np.array([0, 1, 2, 3]))

    assert len(ca_pos) == 0

    assert set(chains) & set(['A']*nresidues)
    assert set(resnames) & set(['res0', 'res1', 'res2', 'res3'])

    assert equal_graphs(make_residue_graph(test_molecule), G)


def test_calculate_overlap(test_molecule):
    # test the overlap is calculated correctly

    result = contact_map._contact_info(test_molecule)
    points = result[2]
    tree = KDTree(points)
    vdw_list = [7] * len(points)
    natoms = len(points)
    vdw_max = 20
    alpha = 1

    overlaps = contact_map._calculate_overlap(tree, vdw_list, natoms, vdw_max, alpha)

    expected = np.array([[0., 1., 1., 1., 1., 1., 0., 0., 0.],
                         [1., 0., 1., 1., 1., 1., 0., 0., 0.],
                         [1., 1., 0., 1., 1., 0., 0., 0., 0.],
                         [1., 1., 1., 0., 1., 1., 1., 1., 0.],
                         [1., 1., 1., 1., 0., 1., 1., 1., 1.],
                         [1., 1., 0., 1., 1., 0., 1., 1., 1.],
                         [0., 0., 0., 1., 1., 1., 0., 1., 1.],
                         [0., 0., 0., 1., 1., 1., 1., 0., 1.],
                         [0., 0., 0., 0., 1., 1., 1., 1., 0.]])

    assert np.allclose(overlaps, expected)


def test_calculate_csu(test_molecule):
    # test that the csu contacts are found correctly

    result = contact_map._contact_info(test_molecule)
    points = result[2]

    vdw_list = [7] * len(points)
    fiba, fibb = 13, 21
    natoms = len(points)
    tree = KDTree(points)
    vdw_max = 20
    water_radius = 1

    csu_contacts = contact_map._calculate_csu(points,
                                              vdw_list,
                                              fiba,
                                              fibb,
                                              natoms,
                                              tree,
                                              vdw_max,
                                              water_radius)

    # these are the values that sphere points might have contacts with for each point above.
    expected = np.array([[-1,  1,  3, -1,  1, -1,  1,  3, -1,  1,  3,  1,  3, -1,  1,  3, -1,  1, -1,  1,  3],
                         [-1,  2,  0, -1,  2,  0,  2,  4,  0,  2,  0,  2,  4,  0,  2,  0, -1,  2,  0,  2,  4],
                         [-1, -1,  1, -1, -1,  1, -1,  4,  1, -1,  1, -1,  4,  1, -1,  1, -1, -1,  1, -1, -1],
                         [-1,  4,  5,  0,  4, -1,  0,  5,  0,  4,  5,  0,  5,  0,  4,  5, 0,  4, -1,  0,  5],
                         [-1, -1,  3,  1,  7,  3,  1,  5,  1,  8,  3,  1,  5,  1,  2,  3, 1,  8,  3,  1, -1],
                         [-1, -1,  6,  3,  6, -1,  3,  6,  3,  6,  6,  3,  6,  3,  4,  6, 3,  6, -1,  3,  6],
                         [-1,  7, -1,  5,  7, -1,  5, -1,  5,  7, -1,  5, -1,  5,  7, -1, 5,  7, -1,  5, -1],
                         [-1,  8,  6,  4,  8,  6,  8, -1,  6,  8,  6,  8, -1,  6,  8,  6, 5,  8,  6,  8, -1],
                         [-1, -1,  7, -1, -1,  7, -1, -1,  7, -1,  7, -1, -1,  7, -1,  7, 4, -1,  7, -1, -1]])

    assert np.allclose(csu_contacts, expected)

def test_contact_types(test_molecule):

    result = contact_map._contact_info(test_molecule)
    points = result[2]

    vdw_list = [7] * len(points)
    fiba, fibb = 13, 21
    natoms = len(points)
    tree = KDTree(points)
    vdw_max = 20
    water_radius = 1

    hits = contact_map._calculate_csu(points,
                                      vdw_list,
                                      fiba,
                                      fibb,
                                      natoms,
                                      tree,
                                      vdw_max,
                                      water_radius)


    natoms = len(hits)
    # this is slightly dodgy, but we get away with it because it's < 10. gives some variety at least
    atypes = np.arange(natoms)

    expected = [np.array([[0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 8., 0., 3., 0., 0., 0., 0.],
                          [0., 7., 0., 0., 2., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 5., 6., 0., 0., 0.],
                          [0., 7., 1., 5., 0., 2., 0., 1., 2.],
                          [0., 0., 0., 7., 1., 0., 9., 0., 0.],
                          [0., 0., 0., 0., 0., 7., 0., 5., 0.],
                          [0., 0., 0., 0., 1., 1., 7., 0., 8.],
                          [0., 0., 0., 0., 1., 0., 0., 7., 0.]]),
                np.array([[0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 8., 0., 0., 0., 0., 0., 0.],
                          [0., 7., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 2., 0., 0., 0.],
                          [0., 0., 0., 0., 1., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0.]]),
                np.array([[0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 3., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 2., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 5., 6., 0., 0., 0.],
                          [0., 7., 1., 5., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 7., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0., 0., 0., 0., 0.]])]

    contact_arrays = contact_map._contact_types(hits, natoms, atypes)

    for i, j in enumerate(contact_arrays):
        assert np.allclose(j, expected[i])


def test_calculate_contacts(test_molecule):

    result = contact_map._contact_info(test_molecule)
    points = result[2]
    res_serial = result[3]
    nresidues = result[9]

    vdw_list = [7] * len(points)
    atypes = np.arange(len(points))

    expected = [np.array([[0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 0., 0., 0.]]),
                np.array([[449.,  66.,   0.,   0.],
                          [227., 287., 146.,   0.],
                          [  0., 227.,   0., 226.],
                          [  0.,   0., 227., 817.]]),
                np.array([[449.,   0.,   0.,   0.],
                          [  0.,   0.,   0.,   0.],
                          [  0.,   0.,   0.,   0.],
                          [  0.,   0.,   0.,   0.]]),
                np.array([[  0.,  66.,   0.,   0.],
                          [227., 287., 146.,   0.],
                          [  0., 227.,   0.,   0.],
                          [  0.,   0.,   0.,   0.]])]


    contact_arrays = contact_map._calculate_contacts(vdw_list, atypes, points, res_serial, nresidues)

    for i, j in enumerate(contact_arrays):
        assert np.allclose(j, expected[i])

def test_get_contacts(test_molecule):

    result = contact_map._contact_info(test_molecule)
    points = result[2]
    res_serial = result[3]
    res_idx = result[7]
    nresidues = result[9]
    molecule_graph = result[10]

    vdw_list = [7] * len(points)
    atypes = np.arange(len(points))

    overlaps, contacts, stabilisers, destabilisers = contact_map._calculate_contacts(vdw_list, atypes, points,
                                                                                     res_serial, nresidues)

    # add something interesting to overlaps otherwise we get nothing
    overlaps = np.array([[0., 1., 0., 0.],
                         [0., 1., 1., 0.],
                         [0., 1., 1., 0.],
                         [0., 0., 1., 0.]])

    contacts_list, all_contacts = contact_map._get_contacts(nresidues, overlaps, contacts, stabilisers, destabilisers,
                                                            res_idx, molecule_graph)

    expected_all_contacts = [[1, 2, 0, 1, 1.0, 66.0, 0.0, False],
                             [2, 1, 1, 0, 0.0, 227.0, 0.0, False],
                             [2, 3, 1, 2, 1.0, 146.0, 0.0, False],
                             [3, 2, 2, 1, 1.0, 227.0, 0.0, False],
                             [3, 4, 2, 3, 0.0, 226.0, 0.0, False],
                             [4, 3, 3, 2, 1.0, 227.0, 0.0, False]]

    expected_contacts_list = [[1, "A", 2, "A"],
                              [2, "A", 3, "A"],
                              [3, "A", 2, "A"],
                              [4, "A", 3, "A"]]
    for i, j in zip([contacts_list, all_contacts], [expected_contacts_list, expected_all_contacts]):
        for k, l in zip(i, j):
            assert list(k) == list(l)

def test_write_contacts(test_molecule, tmp_path):

    result = contact_map._contact_info(test_molecule)
    points = result[2]
    res_serial = result[3]
    res_idx = result[7]
    nresidues = result[9]
    molecule_graph = result[10]
    vdw_list = [7] * len(points)
    atypes = np.arange(len(points))

    # make some fake ca positions from the COGs of the residues
    ca_pos = []
    for residue in molecule_graph.nodes:
        subgraph = molecule_graph.nodes[residue]['graph']
        pos = []
        for atom in sorted(subgraph.nodes):
            pos.append(subgraph.nodes[atom]['position']*10)
        ca_pos.append(np.mean(np.stack(pos), axis=0))

    overlaps, contacts, stabilisers, destabilisers = contact_map._calculate_contacts(vdw_list, atypes, points,
                                                                                     res_serial, nresidues)

    _, all_contacts = contact_map._get_contacts(nresidues, overlaps, contacts, stabilisers, destabilisers,
                                                res_idx, molecule_graph)

    with open(TEST_MOLECULE_CONTACT_MAP) as expectedfile:
        expected_lines = expectedfile.readlines()

    outpath = tmp_path / 'contacts.out'

    contact_map._write_contacts(outpath,
                                all_contacts,
                                ca_pos,
                                molecule_graph)
    DeferredFileWriter().write()

    with open(str(outpath)) as infile:
        written_lines = infile.readlines()

    # skip the first line here because it's the vermouth version
    for line, expected_line in zip(written_lines[1:], expected_lines[1:]):
        assert line == expected_line


@pytest.mark.parametrize('write_out',
                         (False, True)
                         )
def test_do_contacts(test_molecule, tmp_path, write_out):

    if write_out:
        outpath = Path(tmp_path / 'contacts.out')
    else:
        outpath = False

    contacts = contact_map.do_contacts(test_molecule, outpath)

    # because of the vdw radii we actually expect no contacts
    assert len(contacts) == len([]) == 0

    if write_out:
        DeferredFileWriter().write()

        with open(TEST_MOLECULE_CONTACT_MAP) as expectedfile:
            expected_lines = expectedfile.readlines()

        with open(str(outpath)) as infile:
            written_lines = infile.readlines()

        # skip the first line here because it's the vermouth version
        # this will actually just check that the header is written because no contacts are found
        for line, expected_line in zip(written_lines[1:], expected_lines[1:]):
            assert line == expected_line

def test_processor(test_molecule):
    # these don't actually matter because the processor only need coordinates
    atypes = {0: "P1", 1: "SN4a", 2: "SN4a",
              3: "SP1", 4: "C1",
              5: "TP1",
              6: "P1", 7: "SN3a", 8: "SP4"}
    secstruc = {1: "H", 2: "H", 3: "H", 4: "H"}

    system = create_sys_all_attrs(test_molecule,
                                  moltype="molecule_0",
                                  secstruc=secstruc,
                                  defaults={"chain": "A"},
                                  attrs={"atype": atypes})
    assert len(system.go_params["go_map"]) == 0

    processor = GenerateContactMap(write_file=False)
    processor.run_system(system)

    # really just matters that after the processor has been run a list has been added to the dictionary
    assert len(system.go_params["go_map"]) == 1




# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import DATA_PATH
from ..citation_parser import read_bib


def _read_quote_file(filehandle):
    """
    Iterates over `filehandle`, and yields all strings that are not empty.

    Parameters
    ----------
    filehandle: collections.abc.Iterable[str]
        A file opened for reading.

    Yields
    ------
    str
        All stripped elements of `filehandle` that are not empty.
    """
    for line in filehandle:
        line = line.strip()
        if line:
            yield line


with open(DATA_PATH/'citations.bib') as citation_file:
    COMMON_CITATIONS = read_bib(citation_file)

with open(DATA_PATH/'quotes.txt') as quotes_file:
    QUOTES = list(_read_quote_file(quotes_file))
